EXPORTS
    RegisterExtensionFunctions
    double_cmp
    int_cmp
    map_destroy
    map_insert
    map_iterate
    map_make
    node_destroy
    node_insert
    node_iterate
    print_elem
    sqlite3_api
    sqlite3_extension_init
    xcalloc
    xfree
