#ifndef INC_sqlite3TokenTypes_hpp_
#define INC_sqlite3TokenTypes_hpp_

/* $ANTLR 2.7.7 (20171109): "sqlite3.g" -> "sqlite3TokenTypes.hpp"$ */

#ifndef CUSTOM_API
# define CUSTOM_API
#endif

#ifdef __cplusplus
struct CUSTOM_API sqlite3TokenTypes {
#endif
	enum {
		EOF_ = 1,
		ABORT = 4,
		ACTION = 5,
		AUTOINCREMENT = 6,
		AS = 7,
		ASC = 8,
		AND = 9,
		OR = 10,
		BETWEEN = 11,
		CASCADE = 12,
		CASE_T = 13,
		CAST = 14,
		CHECK = 15,
		CREATE = 16,
		COLLATE = 17,
		CONFLICT = 18,
		CONSTRAINT = 19,
		CURRENT_TIME = 20,
		CURRENT_DATE = 21,
		CURRENT_TIMESTAMP = 22,
		DEFAULT = 23,
		DEFERRABLE = 24,
		DEFERRED = 25,
		DELETE = 26,
		DESC = 27,
		ELSE_T = 28,
		END = 29,
		ESCAPE = 30,
		FAIL = 31,
		FILTER = 32,
		FOLLOWING = 33,
		FOREIGN = 34,
		GLOB = 35,
		KEY = 36,
		LIKE = 37,
		TABLE = 38,
		IF_T = 39,
		IGNORE = 40,
		IN = 41,
		INDEX = 42,
		INITIALLY = 43,
		INSERT = 44,
		IMMEDIATE = 45,
		IS = 46,
		NO = 47,
		NOT = 48,
		NULL_T = 49,
		MATCH = 50,
		EXISTS = 51,
		ON = 52,
		OVER = 53,
		PARTITION = 54,
		PRECEDING = 55,
		PRIMARY = 56,
		RAISE = 57,
		RANGE = 58,
		REFERENCES = 59,
		REGEXP = 60,
		REPLACE = 61,
		RESTRICT = 62,
		ROLLBACK = 63,
		ROWID = 64,
		ROWS = 65,
		SET = 66,
		TEMPORARY = 67,
		TEMP = 68,
		THEN = 69,
		UNBOUNDED = 70,
		UNIQUE = 71,
		UPDATE = 72,
		USING = 73,
		VIRTUAL = 74,
		WHEN = 75,
		WHERE = 76,
		WITHOUT = 77,
		TYPE_NAME = 78,
		COLUMNDEF = 79,
		COLUMNCONSTRAINT = 80,
		TABLECONSTRAINT = 81,
		CREATETABLE = 82,
		CREATEINDEX = 83,
		INDEXEDCOLUMN = 84,
		KEYWORDASTABLENAME = 85,
		KEYWORDASCOLUMNNAME = 86,
		DIGIT = 87,
		DOT = 88,
		ID = 89,
		QUOTEDID = 90,
		QUOTEDLITERAL = 91,
		NUMERIC = 92,
		NL = 93,
		COMMENT = 94,
		WS = 95,
		STRINGLITERAL = 96,
		LPAREN = 97,
		RPAREN = 98,
		COMMA = 99,
		SEMI = 100,
		PLUS = 101,
		MINUS = 102,
		STAR = 103,
		TILDE = 104,
		AMPERSAND = 105,
		BITOR = 106,
		OROP = 107,
		EQUAL = 108,
		EQUAL2 = 109,
		GREATER = 110,
		GREATEREQUAL = 111,
		LOWER = 112,
		LOWEREQUAL = 113,
		UNEQUAL = 114,
		UNEQUAL2 = 115,
		BITWISELEFT = 116,
		BITWISERIGHT = 117,
		SELECT = 118,
		SLASH = 119,
		PERCENT = 120,
		NULL_TREE_LOOKAHEAD = 3
	};
#ifdef __cplusplus
};
#endif
#endif /*INC_sqlite3TokenTypes_hpp_*/
